#include "lexer.hpp"

#include <cstring>
#include <iostream>
#include <regex>

//docra`Chunk`+

//++`::Chunk`
docra::Chunk::Chunk()
: kind(INVALID), lineno(1), depth(0)
{}

//docra`Tag`+

//++`.xref_id`
std::string docra::Tag::xref_id() const
{
  static std::regex deny("[^_A-Za-z0-9]+");
  return std::regex_replace("dcr_" + prefix + "_" + symbol + "_" + suffix, deny, "_");
}

docra::Tag docra::operator + (const docra::Tag &lhs, const docra::Tag &rhs)
{
  docra::Tag ret;
  ret.prefix = lhs.prefix;
  ret.symbol = lhs.symbol + rhs.symbol;
  ret.suffix = rhs.suffix;
  return ret;
}

//docra`Token`+

//++`.empty_str`
bool docra::Token::empty_str() const
{
  for (auto &c : str) {
    if (!isspace(c)) {
      return false;
    }
  }
  return true;
}

//docra`Lexer`+
docra::Lexer::Lexer()
: file(NULL), eof(false), scanner(NULL), depth(0), depth_delta(0)
{}

docra::Lexer::~Lexer()
{
}

//++`.lex_eof` Is called by the lower-level lexers.
int docra::Lexer::lex_eof()
{
  if (!eof) {
    eof = true;
    if (chunk.kind != docra::Chunk::INVALID) {
      return chunk.kind;
    }
  }
  chunk.kind = docra::Chunk::INVALID;
  chunk.depth = 0;
  return EOF;
}

bool docra::Lexer::next_token() //++`.next_token`
{
  //? The last token can be accessed at `Lexer.token` (or with `Lexer.current_token` ). _

  //? Tokens are matched using C++11 regular expressions
  // Two regexes are used, the first one is: <?
  static std::regex re_comment(
   "(<\\?)" "|"   //? See `Token::KIND.BLOCK_ENTER`
   "(\\?>)" "|"   //? See `Token::KIND.BLOCK_EXIT`
   "``([^`]+)``" "|" //? See `Token::KIND.LITERAL`
   "(\r?\n)" "|"  //? See `Token::KIND.EOL`
   "(" //? See `Token::KIND.TAG`
    "(([^\\s`]*)`([^`]*)`([^\\s:]*))(:)?"
   ")"
  );
  //?>Tags are formatted using the following format: _
  //? ``prefix{backtick}id{backtick}suffix:`` _
  //? The id can be empty, the suffix and colon can be omitted.
  //? The full syntax is explained in (TODO). _
  //? The second regex is: <?
  static std::regex re_comment_boc(
   "(<\\?)" "|"   //? See `Token::KIND.BLOCK_ENTER`
   "(\\?>)" "|"   //? See `Token::KIND.BLOCK_ENTER`
   "``([^`]*)``" "|" //? See `Token::KIND.LITERAL`
   "(\r?\n)" "|"  //? See `Token::KIND.EOL`
   "(" //? See `Token::KIND.TAG`
    "(([^\\s`]*)`([^`]*)`([^\\s:]*))(:)?" "|"
    "^(([A-Z]+):)" //TODO: get rid of this exception! It only creates problems!
   ")"
  ); //?>This regex is only used at the beginning of a line, the only difference is that an additional tag syntax is allowed.
  //? This syntax matches tags in the form _
  //?  ABCD: _
  //? In this case, ABCD becomes the prefix, the (TODO) id is empty and the colon is not optional anymore. This syntax is only used to mark positions in the code as TODOs, FIXMEs and so forth
  //? For more details, (TODO).
  enum {
    RE_MATCH = 0,
    RE_BLOCK_ENTER,
    RE_BLOCK_EXIT,
    RE_LITERAL,
    RE_EOL,
    RE_TAG,
    RE_TAG_1, RE_TAG_1_PREFIX, RE_TAG_1_SYMBOL, RE_TAG_1_SUFFIX, RE_TAG_1_COLON,
    RE_TAG_2, RE_TAG_2_PREFIX
  };

  if (chunk.kind == Chunk::INVALID && !eof) {
    if (!next_chunk()) {
      return false;
    }
  }

  //#`scoping`
  // Enter or exit tokens are emitted until `token.depth` equals the depth of the last chunk created in `next_chunk` .
  // This implies that depth information is always stored and generated in the low-level lexers.
  if (token.depth != chunk.depth) {
    if (token.depth < chunk.depth) {
      ++token.depth;
      token.kind = Token::SCOPE_ENTER;
      token.lineno = chunk.lineno;
      return true;
    } else if (token.depth > chunk.depth) {
      --token.depth;
      token.kind = Token::SCOPE_EXIT;
      token.lineno = chunk.lineno;
      return true;
    }
  }

  //? It can happen that `next_chunk` returns 'true' and the last chunk is invalid, this is in fact what is returned on eof! See `lex_eof` .
  if (chunk.kind == Chunk::INVALID && eof) {
    return false;
  }

  //? Code chunks are passed on to the parser as code tokens with no modification, only comments are parsed.
  if (chunk.kind == Chunk::CODE) {
    size_t eol = chunk.str.find_first_of("\r\n");
    if (eol == 0) {
      token.kind = Token::EOL;
      token.str = chunk.str;
      token.lineno = chunk.lineno;
      next_chunk();
      return true;
    } else if (eol != std::string::npos) {
      token.kind = Token::CODE;
      token.str = chunk.str.substr(0, eol);
      token.lineno = chunk.lineno;
      chunk.str = chunk.str.substr(eol);
      return true;
    } else {
      token.kind = Token::CODE;
      token.str = chunk.str;
      token.lineno = chunk.lineno;
      next_chunk();
      return true;
    }
  } else if (chunk.kind == Chunk::TAG) {
    token.kind = Token::TAG;
    token.tag = chunk.tag;
    token.lineno = chunk.lineno;
    token.boc = true;
    token.extra = std::move(chunk.extra);
    next_chunk();
    return true;
  }
  // Else it's necessarily Chunk::COMMENT

matched:

  bool prev_boc = token.boc;
  //? Some rules only apply at the beginning of a comment.
  //? In those cases, `.boc` is set accordingly.
  token.boc = (chunk.it == chunk.str.cbegin());

  if (match.size() > 1) {
    if (match[RE_BLOCK_ENTER].matched) {
      token.kind = Token::BLOCK_ENTER;
      token.str = match[RE_BLOCK_ENTER];
    } else if (match[RE_BLOCK_EXIT].matched) {
      token.kind = Token::BLOCK_EXIT;
      token.str = match[RE_BLOCK_EXIT];
    } else if (match[RE_LITERAL].matched) {
      token.kind = Token::LITERAL;
      token.str = match[RE_LITERAL];
    } else if (match[RE_EOL].matched) {
      token.kind = Token::EOL;
      token.str = match[RE_EOL];
    } else if (match[RE_TAG_1].matched) {
      //#`tag`1: A tag is matched using the first syntax in the regex
      token.kind = Token::TAG;
      token.str = match[RE_TAG_1];
      token.tag.prefix = match[RE_TAG_1_PREFIX];
      token.tag.symbol = match[RE_TAG_1_SYMBOL];
      token.tag.suffix = match[RE_TAG_1_SUFFIX];
      token.brief = match.length(RE_TAG_1_COLON) == 0;
    } else if (match[RE_TAG_2].matched) {
      //#`tag`2: A tag is matched using the second syntax in the regex
      token.kind = Token::TAG;
      token.str = match[RE_TAG_2];
      token.tag.prefix = match[RE_TAG_2_PREFIX];
      token.tag.symbol = "";
      token.tag.suffix = "";
    }
    token.lineno = chunk.lineno;
    //? In C++11 regex, anything that comes _after_ the end of the match constitutes the 'suffix'. _
    //? If its 'matched' attribute is true, then it contains a valid value. _
    //? Moving the `Chunk.it` iterator to the beginning of the prefix will move the tokenizer forward. _
    //? If there's nothing left to parse, the tokenizer will consume the next chunk. _
    //? 'true' is returned because one `Token` was already successfully produced. _
    if (match.suffix().matched) {
      chunk.it = match.suffix().first;
    } else {
      next_chunk();
    }
    //? The only way to reset a match variable seems to be to swap it with a new, local variable that will be cleaned up outside this scope.
    std::smatch empty_match;
    swap(match, empty_match);
    return true;
  }

  // std::match_continuous?
  if ((token.boc
   && std::regex_search(chunk.it, chunk.str.cend(), match, re_comment_boc))
   || std::regex_search(chunk.it, chunk.str.cend(), match, re_comment)) {
    //? If the regex matches somewhere after the beginning of the string, 'prefix' will be set accordingly.
    //? This part of the string will be returned as a simple `DOC_STR` token. _
    if (match.prefix().matched) {
      // Bit of a hack to simplify our life when looking for a leading " _" in a comment
      if (token.kind == Token::TAG && prev_boc == true) {
        token.boc = true;
      }
      token.kind = Token::DOC_STR;
      token.str = match.prefix();
      token.lineno = chunk.lineno;
      chunk.it += match.prefix().length();
      return true;
    } else {
      goto matched;
    }
  } else {
    //? A `DOC_STR` is also returned if nothing is matched. _
    if (token.kind == Token::TAG && prev_boc == true) {
      token.boc = true;
    }
    token.kind = Token::DOC_STR;
    token.str.assign(chunk.it, chunk.str.cend());
    next_chunk();
    token.lineno = chunk.lineno;
    return true;
  }
}
