#include "project.hpp"

#include "docra.hpp"

#include <iostream>
#include <sstream>

//ug`config.patterns`1 When patterns are used to specify include and exclude filters, they are processed in the order they appear in the configuration file. The first matching pattern is used to decide whether to include or exclude a given path.

static bool filter_path(const docra::Filters &filters, const fs::path &path, bool default_choice = false)
{
  for (auto &filter : filters) {
    if (std::regex_match(path.string(), filter.second)) {
      return filter.first;
    }
  }
  return default_choice;
}

//docra`Project`+

//docra`Project`
docra::Project::Project(const docra::Config &cfg)
: _cfg(cfg)
{}

//++`::analyze_directory`
bool docra::Project::analyze_directory(int verb, const fs::path &path, int level)
{
  //TODO: very inefficient to redo every time
  int plen = 0;
  fs::path source_path;
  for (auto it = path.begin(); it != path.end(); ++it, ++plen) {}
  for (auto it = path.begin(); it != path.end(); ++it, --plen) {
    if (plen <= 0 || plen <= level) {
      source_path /= *it;
    }
  }
  source_path += fs::path::preferred_separator;

  bool ret = true;
  if (verb >= Config::VERB_INTERNAL) {
    std::cerr << "analyze_directory(" << path << ") = " << source_path << std::endl;
  }
  if (!fs::is_directory(path)) {
    return false;
  }
  bool included = false;

//ug`config.syntax`common.include _ include = _pattern_::
// Include rules determine whether a file or directory that would have otherwise been excluded should instead be analyzed by docra.

//ug`config.syntax`common.exclude _ exclude = _pattern_::
// Exclude rules determine whether a file or directory will be ignored by docra while traversing a given filesystem hierarchy.

  if (!source_path.has_relative_path()) {
    // The source path itself is always included (to ensure it can be traversed despite the exclude rules)
    included = true;
  } else {
    included = filter_path(_cfg.common_filters, source_path, true);
    if (verb >= Config::VERB_DEBUG) {
      std::cerr << source_path << (included ? " included" : " excluded") << std::endl;
    }
  }
  for (auto &dir_entry : fs::directory_iterator(path)) {
    if (included && fs::is_regular_file(dir_entry)) {
      ret &= analyze_file(verb, dir_entry, level + 1);
    }
    if (fs::is_directory(dir_entry)) {
      ret &= analyze_directory(verb, dir_entry, level + 1);
    }
  }
  return ret;
}

//++`::analyze_file`
bool docra::Project::analyze_file(int verb, const fs::path &path, int level)
{
  //TODO: very inefficient to redo every time
  int plen = 0;
  fs::path source_path;
  for (auto it = path.begin(); it != path.end(); ++it, ++plen) {}
  for (auto it = path.begin(); it != path.end(); ++it, --plen) {
    if (plen <= 0 || plen <= level) {
      source_path /= *it;
    }
  }
  if (verb >= Config::VERB_INTERNAL) {
    std::cerr << "analyze_file(" << path << ") = " << source_path << std::endl;
  }
  if (!fs::is_regular_file(path)) {
    return false;
  }
  bool included = filter_path(_cfg.common_filters, source_path, true);
  if (verb >= Config::VERB_DEBUG) {
    std::cerr << source_path << (included ? " included" : " excluded") << std::endl;
  }
  if (!included) {
    return true;
  }
  if (_cfg.cli_pathsonly) {
    std::cout << source_path.string() << std::endl;
    return true;
  }

#ifdef _MSC_VER
  FILE *pfile = _wfopen(path.c_str(), L"r");
#else
  FILE *pfile = fopen(path.c_str(), "r");
#endif
  if (pfile == nullptr) {
    std::cerr << "unable to open file " << path << std::endl;
    return false;
  }
  bool parse = false;
  if (filter_path(_cfg.lex_clike_filters, path)) {
    if (verb >= Config::VERB_DEBUG) {
      std::cerr << source_path << " uses clike lexer" << std::endl;
    }
    docra::Lexer_CLike lexer;
    lexer.init(path, pfile);
    docra::Parser_Source parser(lexer, *this);
    parse = parser.parse(verb, source_path, scopes[source_path]);
  } else if (filter_path(_cfg.lex_hash_filters, path)) {
    if (verb >= Config::VERB_DEBUG) {
      std::cerr << source_path << " uses hash lexer" << std::endl;
    }
    docra::Lexer_Hash lexer;
    lexer.init(path, pfile);
    docra::Parser_Source parser(lexer, *this);
    parse = parser.parse(verb, source_path, scopes[source_path]);
  } else if (filter_path(_cfg.lex_vhdl_filters, path)) {
    if (verb >= Config::VERB_DEBUG) {
      std::cerr << source_path << " uses vhdl lexer" << std::endl;
    }
    docra::Lexer_Vhdl lexer;
    lexer.init(path, pfile);
    docra::Parser_Source parser(lexer, *this);
    parse = parser.parse(verb, source_path, scopes[source_path]);
  } else if (filter_path(_cfg.lex_ini_filters, path)) {
    if (verb >= Config::VERB_DEBUG) {
      std::cerr << source_path << " uses ini lexer" << std::endl;
    }
    docra::Lexer_Ini lexer;
    lexer.init(path, pfile);
    docra::Parser_Source parser(lexer, *this);
    parse = parser.parse(verb, source_path, scopes[source_path]);
  } else if (filter_path(_cfg.lex_yacc_filters, path)) {
    if (verb >= Config::VERB_DEBUG) {
      std::cerr << source_path << " uses yacc lexer" << std::endl;
    }
    docra::Lexer_Yacc lexer;
    lexer.init(path, pfile);
    docra::Parser_Source parser(lexer, *this);
    parse = parser.parse(verb, source_path, scopes[source_path]);
  } else if (filter_path(_cfg.lex_xml_filters, path)) {
    if (verb >= Config::VERB_DEBUG) {
      std::cerr << source_path << " uses xml lexer" << std::endl;
    }
    docra::Lexer_Xml lexer;
    lexer.init(path, pfile);
    docra::Parser_Source parser(lexer, *this);
    parse = parser.parse(verb, source_path, scopes[source_path]);
  }
  fclose(pfile);
  return parse;
}

docra::Suffix *docra::Project::insert_tag(const Tag &tag, Token &&token, Scope *scope)
{
  Symbol &symbol = prefixes[tag.prefix].symbols[tag.symbol];
  Suffixes &suffixes = symbol.suffixes;

  //? Find all equivalent suffixes
  auto range = suffixes.equal_range(tag.suffix);
  for (auto it = range.first; it != range.second; ++it) {
    //? If a suffix at the same position already existed
    if (it->second.lineno_from() == token.lineno) {
      //? Update its suffix (WHY!?)
      //if (it->second.scope && it->second.scope->suffix == &(it->second)) {
      //  it->second.scope->suffix = nullptr;
      //}
      //? Update its scope
      it->second.scope = scope;
      return &(it->second);
    }
  }
  auto it = suffixes.insert(std::make_pair(tag.suffix, Suffix(this, tag, std::move(token))));
  it->second.scope = scope;
  symbol.suffixes_unsorted.push_back(it);
  return &(it->second);
}

docra::Symbol *docra::Project::locate_symbol(const std::string &prefix, const std::string &symbol)
{
  auto prefix_it = prefixes.find(prefix);
  if (prefix_it == prefixes.end()) {
    return nullptr;
  }
  auto symbol_it = prefix_it->second.symbols.find(symbol);
  if (symbol_it == prefix_it->second.symbols.end()) {
    return nullptr;
  }
  return &symbol_it->second;
}

docra::Prefix *docra::Project::locate_prefix(const std::string &prefix)
{
  auto prefix_it = prefixes.find(prefix);
  if (prefix_it == prefixes.end()) {
    return nullptr;
  }
  return &prefix_it->second;
}

docra::Symbol *docra::Project::locate_symbol(const Tag &tag)
{
  return locate_symbol(tag.prefix, tag.symbol);
}

docra::Suffix *docra::Project::locate(const std::string &prefix, const std::string &symbol, const std::string &suffix)
{
  Symbol *symptr = locate_symbol(prefix, symbol);
  if (symptr == nullptr) {
    return nullptr;
  }
  auto suffix_it = symptr->suffixes.find(suffix);
  if (suffix_it == symptr->suffixes.end()) {
    return nullptr;
  }
  return &suffix_it->second;
}

docra::Suffix *docra::Project::locate(const Tag &tag)
{
  return locate(tag.prefix, tag.symbol, tag.suffix);
}

static std::string make_link(const std::string &prefix, const std::string &path, const std::string &suffix, int line)
{
  std::stringstream str;
  str << prefix << path << suffix << line << "[" << path << ":" << line << "]";
  return str.str();
}

void docra::Project::debug()
{
  for (auto &prefix : prefixes) {
    for (auto &symbol : prefix.second.symbols) {
      for (auto &suffix : symbol.second.suffixes) {
        std::cerr << std::endl << "=== " << prefix.first << "`" << symbol.first << "`" << suffix.first << std::endl;
        //std::cerr << std::endl << "Location: "
        // << make_link(base_url, suffix.second.path.string(), suffix.second.lineno_code) << std::endl;

        //std::cerr << std::endl << "Brief: " << suffix.second.token_tag.brief << std::endl;
        //std::cerr << std::endl << "Description:";
        std::cerr << std::endl;
        suffix.second.render_description(_cfg, std::cout, suffix.second.desc_short);
        suffix.second.render_description(_cfg, std::cout, suffix.second.desc_long);
        std::cerr << std::endl;
        if (!suffix.second.params.empty()) {
          std::cerr << "Params:" << std::endl << std::endl;
        }
        for (auto &param : suffix.second.params) {
          if (param.kind != Param::RET) std::cerr << "- `" << param.name << "` ";
          else std::cerr << "- ";
          switch (param.kind) {
            case Param::IN:
              std::cerr << "(IN)";
              break;
            case Param::OUT:
              std::cerr << "(OUT)";
              break;
            case Param::INOUT:
              std::cerr << "(INOUT)";
              break;
            case Param::GENERIC:
              std::cerr << "(GENERIC)";
              break;
            case Param::RET:
              std::cerr << "(RET)";
              break;
            case Param::LOCAL:
              std::cerr << "(LOCAL)";
              break;
            default:
              std::cerr << "(UNKNOWN?)";
              break;
          }
          for (auto &token : param.desc) {
            switch (token.kind) {
              case Token::EOL:
                std::cerr << std::endl;
                break;
              case Token::TAG:
                std::cerr << token.tag.prefix << '`' << token.tag.symbol << '`' << token.tag.suffix;
                break;
              case Token::DOC_STR:
                if (token.boc && token.str.length() > 1 && token.str.substr(0, 2) == "? ") {
                  std::cerr << token.str.substr(2);
                } else {
                  std::cerr << token.str;
                }
                break;
              default:
                std::cerr << token.str;
                break;
            }
          }
          std::cerr << std::endl;
        }
        std::cerr << std::endl;
      }
    }
  }
}
